@@ -1,8 +1,8 @@
 from __future__ import annotations
 
+from collections import OrderedDict
 
 from . import h, t
-from .DefaultOrderedDict import DefaultOrderedDict
 
 if t.TYPE_CHECKING:
     from . import refs as r  # pylint: disable=unused-import
@@ -12,31 +12,31 @@ def addDfnPanels(doc: t.SpecT, dfns: list[t.ElementT]) -> None:
     # Constructs "dfn panels" which show all the local references to a term
     atLeastOnePanel = False
     # Gather all the <a href>s together
-    allRefs = DefaultOrderedDict(list)
+    allRefs: OrderedDict[str, list[t.ElementT]] = OrderedDict()
     for a in h.findAll("a", doc):
         href = a.get("href")
         if href is None:
             continue
         if not href.startswith("#"):
             continue
-        allRefs[href[1:]].append(a)
+        allRefs.setdefault(href[1:], []).append(a)
     for dfn in dfns:
         id = dfn.get("id")
         if not id:
             # Something went wrong, bail.
             continue
-        refs = DefaultOrderedDict(list)
-        for link in allRefs[id]:
-            section = h.sectionName(link)
+        refs: OrderedDict[str, list[t.ElementT]] = OrderedDict()
+        for link in allRefs.get(id, []):
+            section = h.sectionName(doc, link)
             if section is not None:
-                refs[section].append(link)
+                refs.setdefault(section, []).append(link)
         if not refs:
             # Just insert a self-link instead
             # unless it already has a self-link, of course
             if h.find(".self-link", dfn) is None:
                 h.appendChild(dfn, h.E.a({"href": "#" + h.escapeUrlFrag(id), "class": "self-link"}))
             continue
-        h.addClass(dfn, "dfn-paneled")
+        h.addClass(doc, dfn, "dfn-paneled")
         atLeastOnePanel = True
         panel = h.E.aside(
             {"class": "dfn-panel", "data-for": id},
@@ -80,17 +80,27 @@ def addDfnPanels(doc: t.SpecT, dfns: list[t.ElementT]) -> None:
         doc.extraStyles["style-dfn-panel"] = dfnPanelStyle
 
 
-def addExternalDfnPanel(
-    termEl: t.ElementT, ref: r.RefWrapper, elsFromHref: dict[str, list[t.ElementT]], doc: t.SpecT
-) -> None:
+def addExternalDfnPanel(termEl: t.ElementT, ref: r.RefWrapper, doc: t.SpecT) -> None:
     # Constructs "dfn panels" which show all the local references to an external term
-    # Gather all the <a href>s together
-    refs = DefaultOrderedDict(list)
-    for el in elsFromHref[ref.url]:
-        section = h.sectionName(el) or "Unnumbered Section"
-        refs[section].append(el)
-    if len(refs) > 0:
-        h.addClass(termEl, "dfn-paneled")
+
+    # Calculate and cache the doc's links,
+    # so I'm not iterating the doc for links constantly.
+    if not doc.cachedLinksFromHref:
+        for a in h.findAll("a", doc):
+            href = a.get("href")
+            if href is None:
+                continue
+            if href.startswith("#"):
+                continue
+            doc.cachedLinksFromHref.setdefault(href, []).append(a)
+
+    # Group the relevant links according to the section they're in.
+    linksBySection: OrderedDict[str, list[t.ElementT]] = OrderedDict()
+    for link in doc.cachedLinksFromHref[ref.url]:
+        section = h.sectionName(doc, link) or "Unnumbered Section"
+        linksBySection.setdefault(section, []).append(link)
+    if linksBySection:
+        h.addClass(doc, termEl, "dfn-paneled")
         _, _, refID = ref.url.partition("#")
         termID = f"term-for-{refID}"
         termEl.set("id", termID)
@@ -101,7 +111,7 @@ def addExternalDfnPanel(
             h.E.b("Referenced in:"),
         )
         ul = h.appendChild(panel, h.E.ul())
-        for text, els in refs.items():
+        for text, els in linksBySection.items():
             li = h.appendChild(ul, h.E.li())
             for i, el in enumerate(els):
                 linkID = el.get("id")
