@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import json
 import os
 import re
@@ -7,14 +9,85 @@ import certifi
 import tenacity
 from json_home_client import Client as APIClient
 
-from .. import config, messages as m
+from .. import config, messages as m, t
+
+if t.TYPE_CHECKING:
+    AnchorsT: t.TypeAlias = defaultdict[str, list[AnchorT]]
+    HeadingGroupT: t.TypeAlias = dict[SpecStatusT, HeadingT]
+    HeadingKeyT: t.TypeAlias = str
+    SpecStatusT: t.TypeAlias = str
+    HeadingsT: t.TypeAlias = dict[HeadingKeyT, list[HeadingKeyT] | HeadingGroupT]
+    AllHeadingsT: t.TypeAlias = dict[str, HeadingsT]
+    SpecsT: t.TypeAlias = dict[str | None, SpecT]
+    MethodsT: t.TypeAlias = defaultdict[str, dict[str, t.Any]]
+    ForsT: t.TypeAlias = defaultdict[str, list[str]]
+
+    class HeadingT(t.TypedDict):
+        url: str
+        number: str
+        text: str
+        spec: str
+
+    RawAnchorT = t.TypedDict(
+        "RawAnchorT",
+        {
+            "name": t.Required[str],
+            "type": t.Required[str],
+            "for": list[str],
+            "section": bool,
+            "title": str,
+            "status": t.Required[str],
+            "normative": bool,
+            "export": bool,
+            "linking_text": list[str],
+            "uri": t.Required[str],
+        },
+        total=False,
+    )
+
+    # Need to use function form due to "for" key
+    # being invalid as a property name
+    AnchorT = t.TypedDict(
+        "AnchorT",
+        {
+            "status": str,
+            "type": str,
+            "spec": str,
+            "shortname": str,
+            "level": int,
+            "export": bool,
+            "normative": bool,
+            "url": str,
+            "for": list[str],
+        },
+    )
+
+    class RawSpecT(t.TypedDict, total=False):
+        name: t.Required[str]
+        short_name: t.Required[str]
+        title: t.Required[str]
+        level: t.Required[int]
+        description: str | None
+        base_uri: str | None
+        draft_uri: str | None
+        abstract: str | None
+
+    class SpecT(t.TypedDict):
+        vshortname: str
+        shortname: str
+        snapshot_url: str | None
+        current_url: str | None
+        title: str
+        description: str | None
+        abstract: str | None
+        level: int | None
 
 
-def progressMessager(index, total):
+def progressMessager(index: int, total: int) -> t.Callable[[], None]:
     return lambda: m.say(f"Downloading data for spec {index}/{total}...")
 
 
-def update(path, dryRun=False):
+def update(path: str, dryRun: bool = False) -> set[str] | None:
     m.say("Downloading anchor data...")
     shepherd = APIClient(
         "https://api.csswg.org/shepherd/",
@@ -23,12 +96,12 @@ def update(path, dryRun=False):
     )
     rawSpecData = dataFromApi(shepherd, "specifications", draft=True)
     if not rawSpecData:
-        return
+        return None
 
-    specs = dict()
-    anchors = defaultdict(list)
-    headings = defaultdict(dict)
-    lastMsgTime = 0
+    specs: SpecsT = dict()
+    anchors: AnchorsT = defaultdict(list)
+    headings: AllHeadingsT = {}
+    lastMsgTime: float = 0
     for i, rawSpec in enumerate(rawSpecData.values(), 1):
         lastMsgTime = config.doEvery(
             s=5,
@@ -37,10 +110,12 @@ def update(path, dryRun=False):
         )
         rawSpec = dataFromApi(shepherd, "specifications", draft=True, anchors=True, spec=rawSpec["name"])
         spec = genSpec(rawSpec)
+        assert spec["vshortname"] is not None
         specs[spec["vshortname"]] = spec
-        specHeadings = headings[spec["vshortname"]]
+        specHeadings: HeadingsT = {}
+        headings[spec["vshortname"]] = specHeadings
 
-        def setStatus(obj, status):
+        def setStatus(obj: RawAnchorT, status: str) -> RawAnchorT:
             obj["status"] = status
             return obj
 
@@ -50,7 +125,8 @@ def update(path, dryRun=False):
         for rawAnchor in rawAnchorData:
             rawAnchor = fixupAnchor(rawAnchor)
             linkingTexts = rawAnchor["linking_text"]
-            if linkingTexts[0] is None:
+            assert linkingTexts is not None
+            if len(linkingTexts) == 0:
                 # Happens if it had no linking text at all originally
                 continue
             if len(linkingTexts) == 1 and linkingTexts[0].strip() == "":
@@ -58,7 +134,7 @@ def update(path, dryRun=False):
                 continue
             if "section" in rawAnchor and rawAnchor["section"] is True:
                 addToHeadings(rawAnchor, specHeadings, spec=spec)
-            if rawAnchor["type"] not in ["heading"]:
+            else:
                 addToAnchors(rawAnchor, anchors, spec=spec)
 
     cleanSpecHeadings(headings)
@@ -75,21 +151,21 @@ def update(path, dryRun=False):
                 f.write(json.dumps(specs, ensure_ascii=False, indent=2, sort_keys=True))
         except Exception as e:
             m.die(f"Couldn't save spec database to disk.\n{e}")
-            return
+            return None
         try:
-            for spec, specHeadings in headings.items():
-                p = os.path.join(path, "headings", f"headings-{spec}.json")
+            for specName, specHeadings in headings.items():
+                p = os.path.join(path, "headings", f"headings-{specName}.json")
                 writtenPaths.add(p)
                 with open(p, "w", encoding="utf-8") as f:
                     f.write(json.dumps(specHeadings, ensure_ascii=False, indent=2, sort_keys=True))
         except Exception as e:
             m.die(f"Couldn't save headings database to disk.\n{e}")
-            return
+            return None
         try:
             writtenPaths.update(writeAnchorsFile(anchors, path))
         except Exception as e:
             m.die(f"Couldn't save anchor database to disk.\n{e}")
-            return
+            return None
         try:
             p = os.path.join(path, "methods.json")
             writtenPaths.add(p)
@@ -97,7 +173,7 @@ def update(path, dryRun=False):
                 f.write(json.dumps(methods, ensure_ascii=False, indent=2, sort_keys=True))
         except Exception as e:
             m.die(f"Couldn't save methods database to disk.\n{e}")
-            return
+            return None
         try:
             p = os.path.join(path, "fors.json")
             writtenPaths.add(p)
@@ -105,14 +181,14 @@ def update(path, dryRun=False):
                 f.write(json.dumps(fors, ensure_ascii=False, indent=2, sort_keys=True))
         except Exception as e:
             m.die(f"Couldn't save fors database to disk.\n{e}")
-            return
+            return None
 
     m.say("Success!")
     return writtenPaths
 
 
 @tenacity.retry(reraise=True, stop=tenacity.stop_after_attempt(3), wait=tenacity.wait_random(1, 2))
-def dataFromApi(api, *args, **kwargs):
+def dataFromApi(api: APIClient, *args: t.Any, **kwargs: t.Any) -> t.JSONT:
     anchorDataContentTypes = [
         "application/json",
         "application/vnd.csswg.shepherd.v1+json",
@@ -128,65 +204,67 @@ def dataFromApi(api, *args, **kwargs):
             "This version of the anchor-data API is no longer supported. Try updating Bikeshed. If the error persists, please report it on GitHub."
         )
     if res.content_type not in anchorDataContentTypes:
-        raise Exception(f"Unrecognized anchor-data content-type '{res.contentType}'.")
+        raise Exception(f"Unrecognized anchor-data content-type '{res.content_type}'.")
     if res.status_code >= 300:
         raise Exception(
             f"Unknown error fetching anchor data; got status {res.status_code} and bytes:\n{data.decode('utf-8')}"
         )
     if isinstance(data, bytes):
         raise Exception(f"Didn't get expected JSON data. Got:\n{data.decode('utf-8')}")
-    return data
+    return t.cast("t.JSONT", data)
 
 
-def linearizeAnchorTree(multiTree, list=None):
-    if list is None:
-        list = []
+def linearizeAnchorTree(multiTree: list, rawAnchors: list[dict[str, t.Any]] | None = None) -> list[RawAnchorT]:
+    if rawAnchors is None:
+        rawAnchors = []
     # Call with multiTree being a list of trees
     for item in multiTree:
-        if item["type"] in config.dfnTypes.union(["dfn", "heading"]):
-            list.append(item)
+        if (item["type"] == "dfn") or ("section" in item and item["section"] is True):
+            rawAnchors.append(item)
         if item.get("children"):
-            linearizeAnchorTree(item["children"], list)
+            linearizeAnchorTree(item["children"], rawAnchors)
             del item["children"]
-    return list
+    return t.cast("list[RawAnchorT]", rawAnchors)
 
 
-def genSpec(rawSpec):
-    spec = {
+def genSpec(rawSpec: RawSpecT) -> SpecT:
+    assert rawSpec["name"] is not None
+    assert rawSpec["short_name"] is not None
+    assert rawSpec["title"] is not None
+    assert rawSpec["description"] is not None
+    spec: SpecT = {
         "vshortname": rawSpec["name"],
-        "shortname": rawSpec.get("short_name"),
+        "shortname": rawSpec["short_name"],
         "snapshot_url": rawSpec.get("base_uri"),
         "current_url": rawSpec.get("draft_uri"),
-        "title": rawSpec.get("title"),
-        "description": rawSpec.get("description"),
-        "work_status": rawSpec.get("work_status"),
-        "working_group": rawSpec.get("working_group"),
-        "domain": rawSpec.get("domain"),
-        "status": rawSpec.get("status"),
+        "title": rawSpec["title"],
+        "description": rawSpec["description"],
         "abstract": rawSpec.get("abstract"),
+        "level": None,
     }
-    if spec["shortname"] is not None and spec["vshortname"].startswith(spec["shortname"]):
+    vShortname = spec["vshortname"]
+    if spec["shortname"] is not None and vShortname.startswith(spec["shortname"]):
         # S = "foo", V = "foo-3"
         # Strip the prefix
-        level = spec["vshortname"][len(spec["shortname"]) :]
+        level = vShortname[len(spec["shortname"]) :]
         if level.startswith("-"):
             level = level[1:]
         if level.isdigit():
             spec["level"] = int(level)
         else:
             spec["level"] = 1
-    elif spec["shortname"] is None and re.match(r"(.*)-(\d+)", spec["vshortname"]):
+    elif spec["shortname"] is None and re.match(r"(.*)-(\d+)", vShortname):
         # S = None, V = "foo-3"
-        match = re.match(r"(.*)-(\d+)", spec["vshortname"])
+        match = t.cast(re.Match, re.match(r"(.*)-(\d+)", vShortname))
         spec["shortname"] = match.group(1)
         spec["level"] = int(match.group(2))
     else:
-        spec["shortname"] = spec["vshortname"]
+        spec["shortname"] = vShortname
         spec["level"] = 1
     return spec
 
 
-def fixupAnchor(anchor):
+def fixupAnchor(anchor: RawAnchorT) -> RawAnchorT:
     """Miscellaneous fixes to the anchors before I start processing"""
 
     # This one issue was annoying
@@ -194,43 +272,52 @@ def fixupAnchor(anchor):
         anchor["title"] = "@import"
 
     # css3-tables has this a bunch, for some strange reason
-    if anchor.get("uri", "").startswith("??"):
-        anchor["uri"] = anchor["uri"][2:]
+    if (anchor.get("uri") or "").startswith("??"):
+        anchor["uri"] = t.cast(str, anchor.get("uri"))[2:]
 
     # If any smart quotes crept in, replace them with ASCII.
-    linkingTexts = anchor.get("linking_text", [anchor.get("title")])
-    for i, t in enumerate(linkingTexts):
-        if t is None:
+    linkingTexts: list[str] = anchor.get("linking_text") or [t.cast(str, anchor.get("title"))]
+    for i, text in enumerate(linkingTexts):
+        if text is None:
             continue
-        if "’" in t or "‘" in t:
-            t = re.sub(r"‘|’", "'", t)
-            linkingTexts[i] = t
-        if "“" in t or "”" in t:
-            t = re.sub(r"“|”", '"', t)
-            linkingTexts[i] = t
+        if "’" in text or "‘" in text:
+            text = re.sub(r"‘|’", "'", text)
+            linkingTexts[i] = text
+        if "“" in text or "”" in text:
+            text = re.sub(r"“|”", '"', text)
+            linkingTexts[i] = text
     anchor["linking_text"] = linkingTexts
 
     # Normalize whitespace to a single space
-    for k, v in list(anchor.items()):
-        if isinstance(v, str):
-            anchor[k] = re.sub(r"\s+", " ", v.strip())
-        elif isinstance(v, list):
-            for k1, v1 in enumerate(v):
-                if isinstance(v1, str):
-                    anchor[k][k1] = re.sub(r"\s+", " ", v1.strip())
+    def strip(s: str) -> str:
+        return re.sub(r"\s+", " ", s.strip())
+
+    anchor["name"] = strip(anchor["name"])
+    anchor["type"] = strip(anchor["type"])
+    anchor["for"] = [strip(x) for x in anchor.get("for", [])]
+    anchor["status"] = strip(anchor["status"])
+    anchor["linking_text"] = [strip(x) for x in anchor.get("linking_text", []) if x is not None]
+    anchor["uri"] = strip(anchor["uri"])
+
     return anchor
 
 
-def addToHeadings(rawAnchor, specHeadings, spec):
+def addToHeadings(rawAnchor: RawAnchorT, specHeadings: HeadingsT, spec: SpecT) -> None:
     uri = rawAnchor["uri"]
+    assert uri is not None
+    if rawAnchor["status"] == "snapshot":
+        baseUrl = spec["snapshot_url"]
+    else:
+        baseUrl = spec["current_url"]
+    assert baseUrl is not None
+    heading: HeadingT = {
+        "url": baseUrl + uri,
+        "number": rawAnchor["name"] if re.match(r"[\d.]+$", rawAnchor["name"]) else "",
+        "text": rawAnchor["title"],
+        "spec": spec["title"],
+    }
     if uri[0] == "#":
         # Either single-page spec, or link on the top page of a multi-page spec
-        heading = {
-            "url": spec["{}_url".format(rawAnchor["status"])] + uri,
-            "number": rawAnchor["name"] if re.match(r"[\d.]+$", rawAnchor["name"]) else "",
-            "text": rawAnchor["title"],
-            "spec": spec["title"],
-        }
         fragment = uri
         shorthand = "/" + fragment
     else:
@@ -251,44 +338,50 @@ def addToHeadings(rawAnchor, specHeadings, spec):
             page = "/" + page
             fragment = "#"
         shorthand = page + fragment
-        heading = {
-            "url": spec["{}_url".format(rawAnchor["status"])] + uri,
-            "number": rawAnchor["name"] if re.match(r"[\d.]+$", rawAnchor["name"]) else "",
-            "text": rawAnchor["title"],
-            "spec": spec["title"],
-        }
     if shorthand not in specHeadings:
         specHeadings[shorthand] = {}
-    specHeadings[shorthand][rawAnchor["status"]] = heading
+    headingGroup = t.cast("HeadingGroupT", specHeadings[shorthand])
+    headingGroup[rawAnchor["status"]] = heading
+    assert rawAnchor["status"] is not None
     if fragment not in specHeadings:
         specHeadings[fragment] = []
+    keyList = t.cast("list[HeadingKeyT]", specHeadings[fragment])
     if shorthand not in specHeadings[fragment]:
-        specHeadings[fragment].append(shorthand)
+        keyList.append(shorthand)
 
 
-def cleanSpecHeadings(headings):
+def cleanSpecHeadings(headings: AllHeadingsT) -> None:
     """Headings data was purposely verbose, assuming collisions even when there wasn't one.
     Want to keep the collision data for multi-page, so I can tell when you request a non-existent page,
     but need to collapse away the collision stuff for single-page."""
     for specHeadings in headings.values():
         for k, v in list(specHeadings.items()):
-            if k[0] == "#" and len(v) == 1 and v[0][0:2] == "/#":
+            if k[0] != "#":
+                # a HeadingGroupT, not a list of HeadingKeyTs
+                continue
+            assert isinstance(v, list)
+            if len(v) == 1 and v[0][0:2] == "/#":
                 # No collision, and this is either a single-page spec or a non-colliding front-page link
                 # Go ahead and collapse them.
                 specHeadings[k] = specHeadings[v[0]]
                 del specHeadings[v[0]]
 
 
-def addToAnchors(rawAnchor, anchors, spec):
-    anchor = {
+def addToAnchors(rawAnchor: RawAnchorT, anchors: AnchorsT, spec: SpecT) -> None:
+    if rawAnchor["status"] == "snapshot":
+        baseUrl = spec["snapshot_url"]
+    else:
+        baseUrl = spec["current_url"]
+    assert baseUrl is not None
+    anchor: AnchorT = {
         "status": rawAnchor["status"],
         "type": rawAnchor["type"],
         "spec": spec["vshortname"],
         "shortname": spec["shortname"],
-        "level": int(spec["level"]),
+        "level": int(spec["level"] or "1"),
         "export": rawAnchor.get("export", False),
         "normative": rawAnchor.get("normative", False),
-        "url": spec["{}_url".format(rawAnchor["status"])] + rawAnchor["uri"],
+        "url": baseUrl + rawAnchor["uri"],
         "for": rawAnchor.get("for", []),
     }
     for text in rawAnchor["linking_text"]:
@@ -298,10 +391,10 @@ def addToAnchors(rawAnchor, anchors, spec):
         anchors[text].append(anchor)
 
 
-def extractMethodData(anchors):
+def extractMethodData(anchors: AnchorsT) -> MethodsT:
     """Compile a db of {argless methods => {argfull method => {args, fors, url, shortname}}"""
 
-    methods = defaultdict(dict)
+    methods: defaultdict[str, dict] = defaultdict(dict)
     for key, anchors_ in anchors.items():
         # Extract the name and arguments
         match = re.match(r"([^(]+)\((.*)\)", key)
@@ -327,24 +420,24 @@ def extractMethodData(anchors):
     return methods
 
 
-def extractForsData(anchors):
+def extractForsData(anchors: AnchorsT) -> ForsT:
     """Compile a db of {for value => dict terms that use that for value}"""
 
-    fors = defaultdict(set)
+    fors: ForsT = defaultdict(list)
     for key, anchors_ in anchors.items():
         for anchor in anchors_:
             for for_ in anchor["for"]:
                 if for_ == "":
                     continue
-                fors[for_].add(key)
+                fors[for_].append(key)
             if not anchor["for"]:
-                fors["/"].add(key)
+                fors["/"].append(key)
     for key, val in list(fors.items()):
-        fors[key] = sorted(val)
+        fors[key] = sorted(set(val))
     return fors
 
 
-def writeAnchorsFile(anchors, path):
+def writeAnchorsFile(anchors: AnchorsT, path: str) -> set[str]:
     """
     Keys may be duplicated.
 
@@ -361,7 +454,7 @@ def writeAnchorsFile(anchors, path):
     - (by itself, ends the segment)
     """
     writtenPaths = set()
-    groupedEntries = defaultdict(dict)
+    groupedEntries: defaultdict[str, dict[str, list[AnchorT]]] = defaultdict(dict)
     for key, entries in anchors.items():
         group = config.groupFromKey(key)
         groupedEntries[group][key] = entries
